/* Talika panel applet: Switch between open windows using icons.
 * (C) 2009-2010 Sinew Software Systems
	 *
 * Authors:  Vinod Kumar Dahiya
 *           Hemant Kumar
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
	 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <gtk/gtk.h>
#include <gio/gio.h>
#include <gio/gdesktopappinfo.h>

#include "tasklist.h"
#include "task.h"
#include "launcher.h"

/*########################## Utility functions ###############################*/

static gchar *
get_process_name_from_pid (gint pid)
{
	if(pid==0)
		return NULL;

	gchar *name=NULL;
	gchar *file;
	file = g_strdup_printf ("/proc/%d/cmdline",pid);

	gchar *contents;
	gsize len;
	if(g_file_get_contents(file,&contents,&len,NULL))
	{
		//every argument of commandline is terminated by null
		//so arguments other than first already discarded
		name = contents;
	}
	
	g_free(file);
	
	return name;
}

static GAppInfo*
talika_launcher_get_app_info_from_id (const gchar *id_name)
{
	gchar *filename;

	//get desktop file from class name first
	gchar *temp = g_strdup_printf ("%s.desktop",id_name);
	filename = g_utf8_strdown (temp,-1);
	g_free (temp);
	
	GAppInfo *app_info = (GAppInfo *)g_desktop_app_info_new (filename);
	g_free(filename);

	return app_info;
}

static void 
do_free_app_info (gpointer data, 
                    gpointer user_data)
{
	GAppInfo *app_info = (GAppInfo *) data;
	g_object_unref (app_info);
}

static GAppInfo*
talika_launcher_get_app_info_from_process_name (const gchar *process_name)
{
	gchar *pname = g_path_get_basename (process_name);

	GAppInfo *app = NULL;
	
	GList *app_list;
	app_list = g_app_info_get_all ();

	GList *l;
	for (l = app_list; l; l = l->next)
	{
		GAppInfo *app_info = l->data;

		const gchar *tmp = g_app_info_get_executable (app_info);
		gchar **splits = g_strsplit (tmp," ",2);

		gchar *ename = g_path_get_basename (splits[0]);
		g_strfreev (splits);

		if(g_ascii_strcasecmp (ename,pname) == 0)
		{			
			app = g_app_info_dup (app_info);
			g_free (ename);
			break;
		}

		g_free (ename);
	}

	g_list_foreach (app_list,do_free_app_info,NULL);
	g_list_free (app_list);
	g_free (pname);
	return app;
}

static GAppInfo*
talika_launcher_get_app_info_from_name (const gchar *name)
{
	gchar *name1 = g_utf8_casefold (name,-1);
	GAppInfo *app = NULL;
	
	GList *app_list;
	app_list = g_app_info_get_all ();

	GList *l;
	for (l = app_list; l; l = l->next)
	{
		GAppInfo *app_info = l->data;

		const gchar *tmp = g_app_info_get_name (app_info);
		gchar *name2 = g_utf8_casefold (tmp,-1);
		
		gint len = strlen(name1);

		if(strncmp (name1,name2,len) == 0)
		{			
			app = g_app_info_dup (app_info);
			g_free (name2);
			break;
		}

		g_free (name2);
	}

	g_list_foreach (app_list,do_free_app_info,NULL);
	g_list_free (app_list);
	g_free (name1);
	return app;
}

static GAppInfo*
talika_launcher_get_app_info (const gchar *name, gint pid)
{	
	//wine apps are not supported, return immediatly
	if( g_ascii_strcasecmp (name,"Wine")==0)
		return NULL;
	
	GAppInfo *app_info =NULL;
	gchar *process_name = get_process_name_from_pid(pid);

	//try to match class name with desktop id
	app_info = talika_launcher_get_app_info_from_id (name);

	//try to match process name with desktop id
	if(!app_info)
	{
		if(process_name)
		{
			app_info = talika_launcher_get_app_info_from_id (process_name);
		}
	}

	//try to match process name with process names
	if(!app_info)
	{
		if(process_name)
		{
			app_info = 
				talika_launcher_get_app_info_from_process_name (process_name);
		}
	}

	//try to match name with names
	if(!app_info)
	{
			app_info = 
				talika_launcher_get_app_info_from_name (name);
	}
	
	if(process_name)
		g_free (process_name);
	
	return app_info;
}

static GdkPixbuf*
talika_get_icon_from_app_info (const GAppInfo* app_info, guint size)
{
	GdkPixbuf *pixbuf=NULL;
	
	if(app_info)
	{
		GIcon *icon = g_app_info_get_icon(G_APP_INFO(app_info));

		if(G_IS_FILE_ICON(icon))
		{
			GFile *icon_file = g_file_icon_get_file (G_FILE_ICON (icon));
			gchar *icon_name = g_file_get_path (icon_file);
			pixbuf = gdk_pixbuf_new_from_file_at_size (icon_name,
			                                           size,
			                                           size,
			                                           NULL);
			g_free (icon_name);
			//g_object_unref (icon_file);
		}
		else if(G_IS_THEMED_ICON(icon))
		{
			const gchar**icon_names =
				(const gchar**)g_themed_icon_get_names(G_THEMED_ICON(icon));

			GtkIconTheme *icon_theme = gtk_icon_theme_get_default ();
			GtkIconInfo *icon_info =
				gtk_icon_theme_choose_icon (icon_theme,icon_names,size,0);

			if(icon_info)
			{
				const gchar *icon_name = gtk_icon_info_get_filename (icon_info);
				
				pixbuf = gdk_pixbuf_new_from_file_at_size (icon_name,
			                                           size,
			                                           size,
			                                           NULL);

				gtk_icon_info_free (icon_info);
			}
		}
	}

	return pixbuf;
}

GdkPixbuf*
talika_get_icon_from_desktop_file(const gchar *name, gint pid, guint size)
{
	GdkPixbuf *pixbuf=NULL;

	GAppInfo *app_info = talika_launcher_get_app_info (name, pid);
	
	if(app_info)
	{
		pixbuf = talika_get_icon_from_app_info (app_info, size);
		g_object_unref (app_info);
	}

	return pixbuf;
}

gchar*
talika_get_spec_from_class_pid (const gchar *name, gint pid)
{
	gchar *desktop_id = NULL;
	gchar *spec = NULL;
	
	GAppInfo *app_info = talika_launcher_get_app_info(name, pid);
	
	if(app_info)
	{
		const gchar *temp = g_app_info_get_id (app_info);
		gint len = strlen(temp);
		desktop_id = g_strndup(temp,len-8);
		
		spec = g_strdup_printf ("%s:%s",desktop_id,name);
		g_print ("%s\n",spec);
		g_print ("\n");

		g_object_unref (app_info);
		g_free (desktop_id);
	}

	return spec;
}


/*############################### Launcher ###################################*/

static GdkPixbuf* 
talika_launcher_get_icon (TalikaLauncher *launcher)
{
	GdkPixbuf *pixbuf = NULL;

	pixbuf = talika_get_icon_from_app_info (launcher->app_info,
	                                        launcher->tasklist->panel_size - 8);
	
	if(launcher->tasklist->mono_icons && pixbuf)
	{
		GdkPixbuf *temp = pixbuf;
		pixbuf = talika_desaturate_icon (temp);
		g_object_unref (temp);
	}

	return pixbuf;
}

static void
talika_launcher_position_menu (GtkMenu * menu,
                           int *x,
                           int *y, gboolean * push_in, gpointer user_data)
{
	TalikaLauncher *task = (TalikaLauncher *) user_data;
	GtkWidget *widget = task->event_box;

	GtkRequisition requisition;
	GdkScreen *screen;
	int menu_x = 0;
	int menu_y = 0;
	int pointer_x;
	int pointer_y;

	screen = gtk_widget_get_screen (widget);

	gtk_widget_size_request (GTK_WIDGET (menu), &requisition);

	gdk_window_get_origin (widget->window, &menu_x, &menu_y);
	gtk_widget_get_pointer (widget, &pointer_x, &pointer_y);

	//works for buttons only, not event box
	menu_x += widget->allocation.x;
	menu_y += widget->allocation.y;

	if (task->tasklist->panel_orient == PANEL_APPLET_ORIENT_UP ||
	    task->tasklist->panel_orient == PANEL_APPLET_ORIENT_DOWN)
	{
		if (gtk_widget_get_direction (GTK_WIDGET (menu)) !=
		    GTK_TEXT_DIR_RTL)
		{
			if (pointer_x < widget->allocation.width &&
			    requisition.width < pointer_x)
				menu_x += MIN (pointer_x,
				               widget->allocation.width -
				               requisition.width);
		}
		else
		{
			menu_x +=
				widget->allocation.width - requisition.width;
			if (pointer_x > 0
			    && pointer_x < widget->allocation.width
			    && pointer_x <
			    widget->allocation.width - requisition.width)
			{
				menu_x -=
					MIN (widget->allocation.width -
					     pointer_x,
					     widget->allocation.width -
					     requisition.width);
			}
		}
		menu_x = MIN (menu_x,
		              gdk_screen_get_width (screen) -
		              requisition.width);

		if (menu_y > gdk_screen_get_height (screen) / 2)
			menu_y -= requisition.height;
		else
			menu_y += widget->allocation.height;
	}
	else
	{
		if (pointer_y < widget->allocation.height &&
		    requisition.height < pointer_y)
			menu_y +=
			MIN (pointer_y,
			     widget->allocation.height -
			     requisition.height);
		menu_y = MIN (menu_y,
		              gdk_screen_get_height (screen) -
		              requisition.height);

		if (menu_x > gdk_screen_get_width (screen) / 2)
			menu_x -= requisition.width;
		else
			menu_x += widget->allocation.width;

	}

	*x = menu_x;
	*y = menu_y;
	*push_in = TRUE;
}

void 
talika_launcher_launch (GtkWidget *menu_item, TalikaLauncher *launcher)
{
	//gtk version older than 2.14 doesn't have GdkAppLaunchContext, so
	//here is a fake replacement
#if GTK_MINOR_VERSION < 14
	GAppLaunchContext *context;
	context = g_app_launch_context_new ();
#else	
	GdkAppLaunchContext *context;
	context = gdk_app_launch_context_new ();
	gdk_app_launch_context_set_screen (context, gtk_widget_get_screen (menu_item));
	gdk_app_launch_context_set_timestamp (context, gtk_get_current_event_time());
#endif
	
	if (g_app_info_launch(G_APP_INFO(launcher->app_info),
	                      NULL,G_APP_LAUNCH_CONTEXT(context),NULL))
	{
		//sucessfull
	}
	else
	{
		//report error
	}

    g_object_unref (context);
}

void 
talika_launcher_unpin (GtkWidget *menu_item, TalikaLauncher *launcher)
{
	gchar *spec = g_strdup_printf ("%s:%s",launcher->desktop_id,
	                               launcher->class_name);
	
	talika_tasklist_remove_launcher (launcher->tasklist, spec);

	g_free (spec);
}

static gboolean
talika_launcher_button_press_event (GtkWidget * widget,
                                    GdkEventButton * event, gpointer data)
{
	TalikaLauncher *launcher = data;


	if (event->button == 1)
	{
		
	}
	else if (event->button == 2)
	{

	}
	else if (event->button == 3)
	{
		gtk_widget_show_all (launcher->action_menu);
		gtk_menu_popup (GTK_MENU (launcher->action_menu),
		                NULL, NULL,
		                talika_launcher_position_menu, launcher,
		                3, gtk_get_current_event_time ());

		return TRUE;
	}
	
	return FALSE;
}

static void 
talika_launcher_update_visible_state (TalikaLauncher * launcher)
{
	GdkPixbuf *pixbuf;

	/*gtk_button_set_relief (GTK_BUTTON (task->button),
	                       task->
	                       tasklist->button_relief ? GTK_RELIEF_NORMAL :
		                       GTK_RELIEF_NONE);*/

	pixbuf = talika_launcher_get_icon (launcher);
	
	gtk_image_set_from_pixbuf (GTK_IMAGE (launcher->image), pixbuf);
	if (pixbuf)
		g_object_unref (pixbuf);

	gtk_widget_queue_resize (GTK_WIDGET (launcher->tasklist->box));
}

void 
do_talika_launcher_update_visible_state (gpointer data, 
                                     gpointer user_data)
{
	TalikaLauncher *launcher = (TalikaLauncher *) data;
	talika_launcher_update_visible_state (launcher);
}

TalikaLauncher*
talika_launcher_new(TalikaTaskList *tasklist, const gchar *spec)
{
	if(!spec)
		return NULL;

	//spec is like desktop_id:class_name
	gchar **tokens = g_strsplit (spec, ":", 2);

	//ill formated specification
	if (tokens[1] == NULL)
	{
		g_strfreev (tokens);
		return NULL;
	}

	//get desktop info, if .desktop exists
	GAppInfo *app_info =
		talika_launcher_get_app_info (tokens[0], 0);
	if(!app_info)
	{
		g_strfreev (tokens);
		return NULL;
	}

	//create new launcher
	TalikaLauncher *launcher = g_new0 (TalikaLauncher, 1);

	launcher->tasklist = tasklist;
		
	//extract desktop id and class name
	launcher->desktop_id = g_strdup (tokens[0]);
	launcher->class_name = g_strdup (tokens[1]);
	
	g_strfreev (tokens);

	launcher->app_info = app_info;
	launcher->exclude = FALSE;

	//create widgets
	launcher->event_box = gtk_button_new ();

	gtk_widget_set_name (launcher->event_box, "tasklist-button");
	gtk_button_set_relief (GTK_BUTTON(launcher->event_box), GTK_RELIEF_NONE);
	
	g_object_add_weak_pointer (G_OBJECT (launcher->event_box),
	                           (void **) &launcher->event_box);
	
	GdkPixbuf *buf = talika_launcher_get_icon (launcher);

	if(buf)
	{
		launcher->image = gtk_image_new_from_pixbuf (buf);
		g_object_unref (buf);
	}
	else
	{
		launcher->image = gtk_image_new ();
	}

	gtk_container_add (GTK_CONTAINER(launcher->event_box),launcher->image);

	gchar *text = g_strdup_printf(_("Launch %s"),
	                              g_app_info_get_name 
	                              (G_APP_INFO(launcher->app_info)));
	gtk_widget_set_tooltip_text (launcher->event_box, text);
	g_free (text);
	
	g_signal_connect (G_OBJECT (launcher->event_box),
	                  "button_press_event",
	                  G_CALLBACK (talika_launcher_button_press_event), launcher);
	g_signal_connect (G_OBJECT (launcher->event_box),
	                  "clicked",
	                  G_CALLBACK (talika_launcher_launch), launcher);

	gtk_widget_show_all (launcher->event_box);

	//create context menu
	launcher->action_menu = gtk_menu_new ();
	g_object_add_weak_pointer (G_OBJECT (launcher->action_menu),
	                           (void **) &launcher->action_menu);

	GtkWidget *menu_item = gtk_menu_item_new_with_mnemonic (_("Unpin"));
	gtk_menu_shell_append (GTK_MENU_SHELL(launcher->action_menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item),
	                  "activate",
	                  G_CALLBACK (talika_launcher_unpin), launcher);

	menu_item = gtk_menu_item_new_with_mnemonic (_("Launch"));
	gtk_menu_shell_append (GTK_MENU_SHELL(launcher->action_menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item),
	                  "activate",
	                  G_CALLBACK (talika_launcher_launch), launcher);

	return launcher;
}

void
talika_launcher_free(TalikaLauncher *launcher)
{
	if (!launcher)
		return;

	if (launcher->event_box)
	{
		g_object_remove_weak_pointer (G_OBJECT (launcher->event_box),
		                              (void **) &launcher->event_box);
		gtk_widget_destroy (launcher->event_box);
		launcher->event_box = NULL;
		launcher->image = NULL;
	}

	if (launcher->action_menu)
	{
		g_object_remove_weak_pointer (G_OBJECT (launcher->action_menu),
		                              (void **) &launcher->action_menu);
		gtk_widget_destroy (launcher->action_menu);
		launcher->action_menu = NULL;
	}

	g_free(launcher->desktop_id);
	g_free(launcher->class_name);

	g_object_unref (launcher->app_info);

	launcher = NULL;
}
