/* Talika panel applet: Switch between open windows using icons.
 * (C) 2009-2010 Sinew Software Systems
	 *
 * Authors:  Vinod Kumar Dahiya
 *           Hemant Kumar
 *
 * NOTE: Code has been taken from libwnck and metacity sources.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
	 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "xutils.h"
#include <string.h>
#include <stdio.h>
#include "inlinepixbufs.h"
#include <X11/extensions/Xcomposite.h>


/*######################### Get icon from X11 ################################*/

void
_wnck_error_trap_push (void)
{
	gdk_error_trap_push ();
}

int
_wnck_error_trap_pop (void)
{
	XSync (gdk_display, False);
	return gdk_error_trap_pop ();
}

/* The icon-reading code is copied
 * from metacity, please sync bugfixes
 */
static gboolean
find_largest_sizes (gulong * data, gulong nitems, int *width, int *height)
{
	*width = 0;
	*height = 0;

	while (nitems > 0)
	{
		int w, h;
		gboolean replace;

		replace = FALSE;

		if (nitems < 3)
			return FALSE;	/* no space for w, h */

		w = data[0];
		h = data[1];

		if (nitems < ((w * h) + 2))
			return FALSE;	/* not enough data */

		*width = MAX (w, *width);
		*height = MAX (h, *height);

		data += (w * h) + 2;
		nitems -= (w * h) + 2;
	}

	return TRUE;
}

static gboolean
find_best_size (gulong * data,
                gulong nitems,
                int ideal_width,
                int ideal_height, int *width, int *height, gulong ** start)
{
	int best_w;
	int best_h;
	gulong *best_start;
	int max_width, max_height;

	*width = 0;
	*height = 0;
	*start = NULL;

	if (!find_largest_sizes (data, nitems, &max_width, &max_height))
		return FALSE;

	if (ideal_width < 0)
		ideal_width = max_width;
	if (ideal_height < 0)
		ideal_height = max_height;

	best_w = 0;
	best_h = 0;
	best_start = NULL;

	while (nitems > 0)
	{
		int w, h;
		gboolean replace;

		replace = FALSE;

		if (nitems < 3)
			return FALSE;	/* no space for w, h */

		w = data[0];
		h = data[1];

		if (nitems < ((w * h) + 2))
			break;	/* not enough data */

		if (best_start == NULL)
		{
			replace = TRUE;
		}
		else
		{
			/* work with averages */
			const int ideal_size =
				(ideal_width + ideal_height) / 2;
			int best_size = (best_w + best_h) / 2;
			int this_size = (w + h) / 2;

			/* larger than desired is always better than smaller */
			if (best_size < ideal_size && this_size >= ideal_size)
				replace = TRUE;
			/* if we have too small, pick anything bigger */
			else if (best_size < ideal_size &&
			         this_size > best_size)
				replace = TRUE;
			/* if we have too large, pick anything smaller
			 * but still >= the ideal
			 */
			else if (best_size > ideal_size &&
			         this_size >= ideal_size &&
			         this_size < best_size)
				replace = TRUE;
		}

		if (replace)
		{
			best_start = data + 2;
			best_w = w;
			best_h = h;
		}

		data += (w * h) + 2;
		nitems -= (w * h) + 2;
	}

	if (best_start)
	{
		*start = best_start;
		*width = best_w;
		*height = best_h;
		return TRUE;
	}
	else
		return FALSE;
}

static void
argbdata_to_pixdata (gulong * argb_data, int len, guchar ** pixdata)
{
	guchar *p;
	int i;

	*pixdata = g_new (guchar, len * 4);
	p = *pixdata;

	/* One could speed this up a lot. */
	i = 0;
	while (i < len)
	{
		guint argb;
		guint rgba;

		argb = argb_data[i];
		rgba = (argb << 8) | (argb >> 24);

		*p = rgba >> 24;
		++p;
		*p = (rgba >> 16) & 0xff;
		++p;
		*p = (rgba >> 8) & 0xff;
		++p;
		*p = rgba & 0xff;
		++p;

		++i;
	}
}

static gboolean
read_rgb_icon (Window xwindow,
               int ideal_width,
               int ideal_height,
               int ideal_mini_width,
               int ideal_mini_height,
               int *width,
               int *height,
               guchar ** pixdata,
               int *mini_width, int *mini_height, guchar ** mini_pixdata)
{
	Atom type;
	int format;
	gulong nitems;
	gulong bytes_after;
	int result, err;
	gulong *data;
	gulong *best;
	int w, h;
	gulong *best_mini;
	int mini_w, mini_h;

	_wnck_error_trap_push ();
	type = None;
	data = NULL;
	result = XGetWindowProperty (gdk_display,
	                             xwindow,
	                             _wnck_atom_get ("_NET_WM_ICON"),
	                             0, G_MAXLONG,
	                             False, XA_CARDINAL, &type, &format,
	                             &nitems, &bytes_after, (void *) &data);

	err = _wnck_error_trap_pop ();

	if (err != Success || result != Success)
		return FALSE;

	if (type != XA_CARDINAL)
	{
		XFree (data);
		return FALSE;
	}

	if (!find_best_size (data, nitems,
	                     ideal_width, ideal_height, &w, &h, &best))
	{
		XFree (data);
		return FALSE;
	}

	if (!find_best_size (data, nitems,
	                     ideal_mini_width, ideal_mini_height,
	                     &mini_w, &mini_h, &best_mini))
	{
		XFree (data);
		return FALSE;
	}

	*width = w;
	*height = h;

	*mini_width = mini_w;
	*mini_height = mini_h;

	argbdata_to_pixdata (best, w * h, pixdata);
	argbdata_to_pixdata (best_mini, mini_w * mini_h, mini_pixdata);

	XFree (data);

	return TRUE;
}

static void
free_pixels (guchar * pixels, gpointer data)
{
	g_free (pixels);
}

static GdkPixbuf *
scaled_from_pixdata (guchar * pixdata, int w, int h, int new_w, int new_h)
{
	GdkPixbuf *src;
	GdkPixbuf *dest;

	src = gdk_pixbuf_new_from_data (pixdata,
	                                GDK_COLORSPACE_RGB,
	                                TRUE,
	                                8, w, h, w * 4, free_pixels, NULL);

	if (src == NULL)
		return NULL;

	if (w != h)
	{
		GdkPixbuf *tmp;
		int size;

		size = MAX (w, h);

		tmp = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, size,
		                      size);

		if (tmp != NULL)
		{
			gdk_pixbuf_fill (tmp, 0);
			gdk_pixbuf_copy_area (src, 0, 0, w, h,
			                      tmp,
			                      (size - w) / 2, (size - h) / 2);

			g_object_unref (src);
			src = tmp;
		}
	}

	if (w != new_w || h != new_h)
	{
		dest = gdk_pixbuf_scale_simple (src, new_w, new_h,
		                                GDK_INTERP_HYPER);

		g_object_unref (G_OBJECT (src));
	}
	else
	{
		dest = src;
	}

	return dest;
}

//get icon from x server, null if failed
GdkPixbuf *
talika_get_icon_from_xserver (WnckWindow * window, guint size)
{
	GdkPixbuf *icon;
	GdkPixbuf *mini_icon;
	guchar *pixdata;
	int w, h;
	guchar *mini_pixdata;
	int mini_w, mini_h;

	gulong xid = wnck_window_get_xid (window);

	icon = NULL;
	mini_icon = NULL;

	if (read_rgb_icon (xid,
	                   size, size,
	                   16, 16,
	                   &w, &h, &pixdata, &mini_w, &mini_h, &mini_pixdata))
	{
		icon = scaled_from_pixdata (pixdata, w, h, size,
		                            size);

		mini_icon = scaled_from_pixdata (mini_pixdata, mini_w, mini_h,
		                                 16,
		                                 16);
		g_object_unref (mini_icon);
		return icon;
	}

	return NULL;
}


/*######################### Get window picture ###############################*/

static GdkPixbuf *
meta_ui_get_pixbuf_from_pixmap (Pixmap   pmap)
{
	GdkPixmap *gpmap;
	GdkScreen *screen;
	GdkPixbuf *pixbuf;
	GdkColormap *cmap;
	int width, height, depth;

	gpmap = gdk_pixmap_foreign_new (pmap);
	screen = gdk_drawable_get_screen (gpmap);

	gdk_drawable_get_size (GDK_DRAWABLE (gpmap), &width, &height);

	depth = gdk_drawable_get_depth (GDK_DRAWABLE (gpmap));
	if (depth <= 24)
		cmap = gdk_screen_get_rgb_colormap (screen);
	else
		cmap = gdk_screen_get_rgba_colormap (screen);

	pixbuf = gdk_pixbuf_get_from_drawable (NULL, gpmap, cmap, 0, 0, 0, 0,
	                                       width, height);

	g_object_unref (gpmap);

	return pixbuf;
}

static GdkPixbuf *
get_window_pixbuf (Display *dpy, Window window, gint size, int *width, int *height)
{
	Pixmap pmap;
	GdkPixbuf *pixbuf, *scaled;
	double ratio;
	double MAX_PREVIEW_SIZE = size;

	gdk_error_trap_push ();
	/* ... Call the X function which may cause an error here ... */
	pmap = XCompositeNameWindowPixmap (dpy, window);
	/* Flush the X queue to catch errors now. */
	gdk_flush ();
	if (gdk_error_trap_pop ())
	{
		/* ... Handle the error here ... */
		pmap = None;
	}

	if (pmap == None||pmap == BadMatch)
		return NULL;

	pixbuf = meta_ui_get_pixbuf_from_pixmap (pmap);
	XFreePixmap(dpy,pmap);

	if (pixbuf == NULL) 
		return NULL;

	*width = gdk_pixbuf_get_width (pixbuf);
	*height = gdk_pixbuf_get_height (pixbuf);

	/* Scale pixbuf to max dimension MAX_PREVIEW_SIZE */
	/*if (*width > *height)
	{
		ratio = ((double) *width) / MAX_PREVIEW_SIZE;
		*width = (int) MAX_PREVIEW_SIZE;
		*height = (int) (((double) *height) / ratio);
}
else*/
	{
		ratio = ((double) *height) / MAX_PREVIEW_SIZE;
		*height = (int) MAX_PREVIEW_SIZE;
		*width = (int) (((double) *width) / ratio);
	}

	scaled = gdk_pixbuf_scale_simple (pixbuf, *width, *height,
	                                  GDK_INTERP_BILINEAR);
	g_object_unref (pixbuf);
	return scaled;
}

GdkPixbuf *
talika_get_window_picture (WnckWindow * window, guint *size)
{
	Display *display = gdk_x11_display_get_xdisplay (
	                                                 gdk_display_get_default ());
	Window win = wnck_window_get_xid (window);

	gint h=0,w=0;

	GdkPixbuf *buf = get_window_pixbuf (display,win,*size,&w,&h);
	*size = w;

	return buf;
}


/*######################### Utility functions ################################*/

//make a GtkLabel bold
void 
_make_gtk_label_bold (GtkLabel * label)
{
	PangoFontDescription *font_desc;

	font_desc = pango_font_description_new ();

	pango_font_description_set_weight (font_desc, PANGO_WEIGHT_BOLD);

	/* This will only affect the weight of the font, the rest is
	 * from the current state of the widget, which comes from the
	 * theme or user prefs, since the font desc only has the
	 * weight flag turned on.
	 */
	gtk_widget_modify_font (GTK_WIDGET (label), font_desc);

	pango_font_description_free (font_desc);
}

//get Gdk screen corsponding to Wnck screen
GdkScreen*
talika_screen_get_gdk_screen (WnckScreen * screen)
{
	g_return_val_if_fail (WNCK_IS_SCREEN (screen), NULL);

	return gdk_display_get_screen (gdk_display_get_default (),
	                               wnck_screen_get_number (screen));
}


/*########################### Preview Cache ##################################*/

static void
talika_preview_image_cache_value_free(gpointer value)
{
	GdkPixbuf *buf = value;
	g_object_unref (buf);
}

TalikaPreviewImageCache* 
talika_preview_image_cache_new(void)
{
	TalikaPreviewImageCache *cache;
	cache = g_new0(TalikaPreviewImageCache,1);
	cache->hash = g_hash_table_new_full (NULL,NULL,NULL,
	                                     talika_preview_image_cache_value_free);
	return cache;
}

void
talika_preview_image_cache_free(TalikaPreviewImageCache *cache)
{
	if(!cache)
		return;
	
	g_hash_table_unref (cache->hash);
}

void
talika_preview_image_cache_update_window (TalikaPreviewImageCache *cache,
                                          WnckWindow *window,
                                          guint preview_size,
                                          GtkWidget *widget)
{
	if(!cache)
		return;
	
	//return if screen not composited
	if( !gdk_screen_is_composited (gtk_widget_get_screen
	                               (widget)) )
		return;

	GdkPixbuf *buf=NULL;
	guint size=preview_size;

	if (!wnck_window_is_minimized (window))
	{
		buf = talika_get_window_picture (window, &size);
	}

	if(buf)
	{
		g_hash_table_replace (cache->hash,window,buf);
	}
}

GdkPixbuf*
talika_preview_image_cache_get_image (TalikaPreviewImageCache *cache,
                                      WnckWindow *window,
                                      guint *size)
{
	if(!cache)
		return NULL;
	
	GdkPixbuf *temp=NULL;
	GdkPixbuf *buf=NULL;

	temp = g_hash_table_lookup (cache->hash,window);

	if(temp)
	{
		*size = gdk_pixbuf_get_width (temp);
		buf = gdk_pixbuf_copy (temp);
	}

	return buf;
}

void
talika_preview_image_cache_free_window (TalikaPreviewImageCache *cache,
                                        WnckWindow *window)
{
	if(!cache)
		return;
	
	g_hash_table_remove (cache->hash,window);
}
